public class ODESpring
{
  //ode stuff
  public int numEqns = 2;
  public float[] q = new float[2]; //q[0] = vel, q[1] = pos
  
  //independent variable
  public float time;

  //spring specifc stuff
  public float mass;
  public float mu;
  public float k;
  public float x0;
  
  public ODESpring(float mass, float mu, float k, float x0)
  {
    this.mass = mass;
    this.mu = mu; //dampening factor 
    this.k = k; //spring constant
    this.x0 = x0; //initial position

    q[0] = 0.0; //initial velocity
    q[1] = x0; //initial position
  }

  public void updatePositionAndVelocity_VI(float dt)
  {
    velocityVerlet(this, dt);
  }

  public void updatePositionAndVelocity_ES(float dt)
  {
    eulerStep(this, dt);
  }

  public void updatePositionAndVelocity_RK(float dt)
  {
    // rungeKutta2(this, dt);
    rungeKutta4(this, dt);  
  }

  public float[] nextSpring( float[] q, float[] deltaQ, float deltaS, float qScale)
  {
    float dq[] = new float[2];
    float newQ[] = new float[2];

    for (int i = 0; i < 2; i++)
    {
      newQ[i] = q[i] + qScale*deltaQ[i];  
    }

    dq[0] = -deltaS * (mu*newQ[0] + k*newQ[1])/mass;
    dq[1] = deltaS*(newQ[0]);

    return dq;
  }


  /*** numerical methods to solve ODE ***/

  //use this version of the Verlet if your model requires velocity
  public void velocityVerlet(ODESpring ode, float dt)
  {
    float[] dq1 = ode.nextSpring(ode.q, ode.q, dt, 0.0);

    float vel = ode.q[0] + (dq1[1] / ode.mass) * dt / 2;
    float pos = ode.q[1] + ode.q[0] * dt;
    float tempq[] = new float[]{
      vel, pos    };

    float[] dq2 = ode.nextSpring( tempq, tempq, dt, 0.0);

    vel = vel + dq2[0] + (dq2[1] / ode.mass) * dt / 2;

    ode.q[1] = pos;
    ode.q[0] = vel;   
  }

  //not being used for springs since model requires velocity!
  public void verletIntegration(ODESpring ode, float dt)
  {
    float prevPos = ode.q[0];
    float curPos = ode.q[1];

    //we need the velocity for the spring, so figure it out...
    ode.q[0] = ode.q[1] - ode.q[0];

    float[] dq1 = ode.nextSpring( ode.q, ode.q, dt, 0.0);

    ode.q[1] = (2 * ode.q[1]) - prevPos - (dt * dt * dq1[1])  ;

    ode.q[0] = curPos; //store current val as previous
  }

  public void eulerStep(ODESpring ode, float dt)
  {
    float[] dq1 = ode.nextSpring(ode.q, ode.q, dt, 0.0);

    for (int i = 0; i < ode.numEqns; i++)
    {
      ode.q[i] += dq1[i]; 
    }    
  }

  public void rungeKutta2(ODESpring ode, float dt)
  {
    //make space to store samples
    float[] dq1 = new float[2];
    float[] dq2 = new float[2];

    //define weight    
    float w1 = 1.0;
    float w2 = 1.0;

    //get estimates of slope at different points
    dq1 = ode.nextSpring( ode.q, ode.q, dt, 0.0);
    dq2 = ode.nextSpring( ode.q, dq1, dt, 1.0);

    //update dependent variable (position)
    for (int i = 0; i < ode.numEqns; i++)
    {
      ode.q[i] += 
        (w1 * dq1[i] + w2 * dq2[i]) / (w1 + w2); 
    }    
  }

  public void rungeKutta4(ODESpring ode, float dt)
  {
    //make space to store samples
    float[] dq1 = new float[2];
    float[] dq2 = new float[2];
    float[] dq3 = new float[2];
    float[] dq4 = new float[2];

    //define weights
    float w1 = 1.0;
    float w2 = 2.0;
    float w3 = 2.0;
    float w4 = 1.0;

    //get estimates of slope at different points
    dq1 = ode.nextSpring( ode.q, ode.q, dt, 0.0);
    dq2 = ode.nextSpring( ode.q, dq1,   dt, 0.5);
    dq3 = ode.nextSpring( ode.q, dq2,   dt, 0.5);
    dq4 = ode.nextSpring( ode.q, dq3,   dt, 1.0);

    //update dependent variable (position)
    for (int i = 0; i < ode.numEqns; i++)
    {
      ode.q[i] += 
        (w1 * dq1[i] + w2 * dq2[i] + w3 * dq3[i] + w4 * dq4[i]) / (w1 + w2 + w3 + w4); 
    }    
  }

}


